@@ -1,3 +1,5 @@
+import base64
+import json
 from lib2to3.pgen2 import token
 import re
 from Crypto.PublicKey import RSA
@@ -95,6 +97,7 @@ def get_service_tokens(self, **kwargs):
     else:
         refresh_token_ttl = kwargs['refresh_token_ttl']
     for tenant_id in self.service_tokens.keys():
+        logger.debug(f"attempting to generate token for tenant {tenant_id}")
         try:
             target_site_id = self.tenant_cache.get_tenant_config(tenant_id=tenant_id).site_id
         except Exception as e:
@@ -117,6 +120,7 @@ def get_service_tokens(self, **kwargs):
                                            f"target_site_id: {target_site_id};\n "
                                            f"request url: {e.request.url}; \n "
                                            f"headers: {e.request.headers}")
+        logger.debug(f"generate token for tenant {tenant_id} successfully.")
         self.service_tokens[tenant_id] = {'access_token': self.add_claims_to_token(tokens.access_token),
                                           'refresh_token': tokens.refresh_token}
 
@@ -218,12 +222,12 @@ def preprocess_service_request(operation, prepared_request, **kwargs):
                                                                                                     operation.resource_name)
 
         elif conf.python_framework_type == 'django':
-            msg = """Did not find `_tapis_use_service_home_site` or `_x_tapis_tenant` and `_x_tapis_user` in kwargs; 
+            msg = """Did not find `_tapis_set_x_headers_from_service` or `_x_tapis_tenant` and `_x_tapis_user` in kwargs; 
             Automatic derivation of these properties is currently only supported for APIs written in flask. 
             If your API is written in flask, be sure to set conf.python_framework_type == flask."""
             raise errors.BaseTapisError(msg)
         else:
-            msg = """Did not find `_tapis_use_service_home_site` or `_x_tapis_tenant` and `_x_tapis_user` in kwargs; 
+            msg = """Did not find `_tapis_set_x_headers_from_service` or `_x_tapis_tenant` and `_x_tapis_user` in kwargs; 
             Automatic derivation of these properties is currently only supported for APIs written in flask. 
             If your API is written in flask, be sure to set conf.python_framework_type == flask."""
             raise errors.BaseTapisError(msg)
@@ -242,7 +246,7 @@ def preprocess_service_request(operation, prepared_request, **kwargs):
     # the tenant_id for the request could be a user tenant (e.g., "tacc" or "dev") but the
     # service tokens are stored by admin tenant, so we need to get the admin tenant for the
     # owning site of the tenant.
-    for tn in operation.tapis_client.tenant_cache.tenants:
+    for tn in operation.tapis_client.tenant_cache.tenants.values(): # tenant_cache is a dict
         if tn.site.site_id == site_id:
             request_site_admin_tenant_id = tn.site.site_admin_tenant_id
     logger.debug(f"site admin tenant for the request: {request_site_admin_tenant_id}")
@@ -263,11 +267,17 @@ def preprocess_service_request(operation, prepared_request, **kwargs):
         # The remaining code is to check whether the token has expired and refresh if necessary
         # if the access token doesn't have an expires_in attribute, there isn't much we can do.
         if not hasattr(access_token, "expires_in"):
-            # it is expected that some tokens, including tokens created by and for the Tokens API itself, will be raw
-            # string types
+            #  tokens created by and for the Tokens API itself do not have an expired_in attr; they are
+            #  service.models.TapisAccessToken type
+            try: 
+                if operation.tapis_client.username == 'tokens':
+                    return
+            except:
+                pass
+            # it is expected that some tokens, including will be raw string types or 
             if not type(access_token) == str:
-                logger.warn(f"The access token didn't have an expired_in attr and was not a string type.")
-            logger.debug("returning from preprocess_service_request")
+                logger.warn(f"The access token didn't have an expired_in attr and was not a string type. type was: {type(access_token)}")
+            logger.debug("returning from preprocess_service_request without checking for the need to refresh.")
             return
         # check the time remaining on the access token ---        
         
@@ -283,7 +293,7 @@ def preprocess_service_request(operation, prepared_request, **kwargs):
                 else:
                     logger.info("service tokens expired, attempting to refresh service tokens.")
                     access_token = operation.tapis_client.refresh_service_tokens(tenant_id=request_site_admin_tenant_id)
-                    logger.info("service tokens refreshed sucessfully.")
+                    logger.info("service tokens refreshed successfully.")
                     prepared_request.headers['X-Tapis-Token']= access_token.access_token
                     # also remove the basic auth header; we shouldn't send both
                     prepared_request.headers.pop('Authorization', None)
@@ -331,7 +341,7 @@ def resolve_tenant_id_for_request(request_thread_local, request, tenant_cache=te
     Additionally, sets the request_thread_local.request_tenant_base_url variable in the process. Returns the
     request_tenant_id (string).
     
-    request should be the request object for the incomming request. at a minimum it should have the following:
+    request should be the request object for the incoming request. at a minimum it should have the following:
       * request.headers - the headers of the request, as a dictionary-like object.
       * request.base_url - the base URL for the request which should include the domain (e.g., localhost or dev.tapis.io).
     request_base_url should be the base URL on the request object. for example, in the flask framework, it is the
@@ -376,10 +386,11 @@ def resolve_tenant_id_for_request(request_thread_local, request, tenant_cache=te
     # the request_baseurl includes the protocol, port (if present) and contains the url path; examples:
     #  http://localhost:5000/v3/oauth2/tenant;
     #  https://dev.develop.tapis.io/v3/oauth2/tenant
-    # in the local development case, the base URL (e.g., localhost:5000...) cannot be used to resolve the tenant id
-    # so instead we use the tenant_id claim within the x-tapis-token:
-    if 'http://172.17.0.1:' in request.base_url or 'http://localhost:' in request.base_url:
-        logger.warn("found 172.17.0.1 or localhost in flask_baseurl; we are assuming this is local development!!")
+    # in the local development case, the base URL (e.g., localhost:5000, 172.17.0.1, dev_base_url conf var)
+    # cannot be used to resolve the tenant id so instead we use the tenant_id claim within the x-tapis-token:
+    dev_request_url = conf.get("dev_request_url", "dev://request_url")
+    if 'http://172.17.0.1:' in request.base_url or 'http://localhost:' in request.base_url or dev_request_url in request.base_url:
+        logger.warn(f"found 172.17.0.1, localhost, or {dev_request_url} in base_url; we are assuming this is local development!!")
         # some services, such as authenticator, have endpoints that do not receive tokens. in the local development
         # case for these endpoints, we don't have a lot of good options -- we can't use the base URL or the token
         # to determine the tenant, so we just set it to the "dev" tenant.
@@ -431,6 +442,10 @@ def validate_request_token(request_thread_local, tenant_cache=tenant_cache):
         logger.error(f"request_thread_local missing token_claims! attrs: {dir(request_thread_local)}")
     request_thread_local.token_claims = claims
     request_thread_local.username = claims.get('tapis/username')
+    # We initialize request_username to look at token username. We overwrite this if it's a service account
+    # later by setting the var equal to request_thread_local.x_tapis_user. Services should use this variable
+    # to process request with correct username.
+    request_thread_local.request_username = claims.get('tapis/username')
     request_thread_local.tenant_id = claims.get('tapis/tenant_id')
     request_thread_local.account_type = claims.get('tapis/account_type')
     request_thread_local.delegation = claims.get('tapis/delegation')
@@ -438,6 +453,8 @@ def validate_request_token(request_thread_local, tenant_cache=tenant_cache):
     if claims.get('tapis/account_type') == 'service':
         request_thread_local.site_id = claims.get('tapis/target_site_id')
         service_token_checks(request_thread_local, claims, tenant_cache)
+        # Overwrite request_username with x_tapis_user which is set using `_x_tapis_user` headers by service
+        request_thread_local.request_username = request_thread_local.x_tapis_user
     # user tokens must *not* set the X-Tapis-Tenant and X-Tapis_user headers
     else:
         try:
@@ -447,6 +464,31 @@ def validate_request_token(request_thread_local, tenant_cache=tenant_cache):
             pass
 
 
+def insecure_decode_jwt_to_claims(token):
+    """
+    Returns the claims associated with a token WITHOUT checking the signature.
+    """
+    # JWT is 3 parts delineated by '.' character
+    parts = token.split('.')
+    # should be three parts
+    if not len(parts) == 3:
+        raise errors.BaseTapisError(f"Invalid JWT format; did not get 3 parts got: {len(parts)}.")
+    # convert part 1 to bytes and add additional padding; see this issue: 
+    # https://stackoverflow.com/questions/2941995/python-ignore-incorrect-padding-error-when-base64-decoding
+    part_bytes = parts[1].encode()
+    part_bytes = part_bytes + b'=='
+    # try to base64 decode the middle part after adding padding; the decode returns `data` as bytes
+    try:
+        data = base64.b64decode(part_bytes)
+    except Exception as e:
+        raise errors.BaseTapisError(f"could not b64 decode the data; exception: {e}")
+    try:
+        claims = json.loads(data)
+    except Exception as e:
+        raise errors.BaseTapisError(f"Could not serialize the bytes data; exception: {e}")
+    return claims
+
+
 def validate_token(token, tenant_cache=tenant_cache):
     """
     Stand-alone function to validate a Tapis token. 
@@ -460,7 +502,7 @@ def validate_token(token, tenant_cache=tenant_cache):
     if not token:
         raise errors.NoTokenError("No Tapis access token found in the request.")
     try:
-        data = jwt.decode(token, verify=False)
+        data = insecure_decode_jwt_to_claims(token)
     except Exception as e:
         logger.debug(f"got exception trying to parse data from the access_token jwt; exception: {e}")
         raise errors.AuthenticationError("Could not parse the Tapis access token.")
@@ -487,7 +529,7 @@ def validate_token(token, tenant_cache=tenant_cache):
     while tries < 2:
         tries = tries + 1
         try:
-            claims = jwt.decode(token, public_key_str)
+            claims = jwt.decode(token, public_key_str, algorithms=["RS256"])
         except Exception as e:
             # if we get an exception decoding it could be that the tenant's public key has changed (i.e., that
             # the public key in out tenant_cache is stale. if we haven't updated the tenant_cache in the last
@@ -550,4 +592,4 @@ def service_token_checks(request_thread_local, claims, tenant_cache):
     # make sure this service is not on the list of services deployed at the associate site --
     if conf.service_name in request_tenant.site.services:
         raise errors.AuthenticationError(f"The primary site does not handle requests to service {conf.service}")
-    logger.debug("this service is NOT in the JWT tenant's owning site's set of services. allowing the request.")
\ No newline at end of file
+    logger.debug("this service is NOT in the JWT tenant's owning site's set of services. allowing the request.")
